/********************************************************************
 * Command Module - processes incoming character data, handles the
 * cursor and contains position logic.
 *
 * Commands:
 *   Displayable Character : load character (at current cursor position)
 *      Converts lower-case letters to upper-case
 *   ^H - Backspace : backspace cursor on same line (no change at position 0)
 *   ^I - Tab : forward space cursor
 *   ^J - Line Feed : move cursor to next line (no change on last line)
 *   ^L - Form Feed : clear display
 *   ^M - Carriage Return : move cursor to position 0 on the current line
 *   ^N - Shift Out : set destination to shadow memory
 *   ^O - Shift In : set destination to display
 *   ^Q - DC1 : enable cursor (applies only when destination is display)
 *   ^R - DC2 : disable cursor (applies only when destination is display)
 *   ^S - DC3 : enable cursor blink (applies only when destination is display)
 *   ^T - DC4 : disable cursor blink (applies only when destination is display)
 *   ^V - SIDLE : Display firmware version
 *   ^Z - Substitute : Load shadow memory
 *   DEL - Delete : Change previous character to a blank and backspace
 *   ^[ - Escape : Set Cursor Position
 *      <ESC>COL;ROW'H'  (special case: <ESC>'H' set to home (0,0)
 *
 *******************************************************************/
#include <xc.h>
#include <stdbool.h>
#include <stdint.h>
#include "cmd.h"
#include "display.h"
#include "system.h"
#include "usb_if.h"

// Command Constants
#define CMD_BS  0x08
#define CMD_TAB 0x09
#define CMD_LF  0x0A
#define CMD_FF  0x0C
#define CMD_CR  0x0D
#define CMD_SO  0x0E
#define CMD_SI  0x0F
#define CMD_DC1 0x11
#define CMD_DC2 0x12
#define CMD_DC3 0x13
#define CMD_DC4 0x14
#define CMD_V   0x16
#define CMD_SUB 0x1A
#define CMD_ESC 0x1B
#define CMD_DEL 0x7F

#define CMD_STATE_IDLE     0
#define CMD_STATE_CUR_COL  1
#define CMD_STATE_CUR_ROW  2


// Cursor Constants
#define CMD_CURSOR_MODE_SOLID 0
#define CMD_CURSOR_MODE_BLINK 1

#define CMD_CURSOR_STATE_NOTVISIBLE  0
#define CMD_CURSOR_STATE_VISIBLE_ON  1
#define CMD_CURSOR_STATE_VISIBLE_OFF 2

// Processing state
uint8_t cmdState;
uint8_t cmdDisplayMask;

// Position Variables
uint8_t cmdDisplayCol;   // Display module Column (0-3)
uint8_t cmdDisplayRow;   // Display module Row (0-15)
uint8_t cmdDisplayX;     // Starting Column for this display

// Cursor Variables
uint8_t cmdCursorEnable;
uint8_t cmdCursorMode;
uint8_t cmdCursorState;
uint8_t cmdCursorCol;    // Cursor Column (0-63)
uint8_t cmdCursorRow;    // Cursor Row (0-15)

uint8_t cmdCursorSaveCol[2]; // Used to swap between memory modes (0: imm, 1:shadow)
uint8_t cmdCursorSaveRow[2];

uint8_t cmdCursorNewCol;  // Used when processing set cursor command
uint8_t cmdCursorNewRow;


// Internal Routine forward declarations
void _CMD_InitPosition();
uint8_t _CMD_PosIsOurs(uint8_t x, uint8_t y);
uint8_t _CMD_MyOffset(uint8_t x);
uint8_t _CMD_CharIsDisplayable(uint8_t* c);
uint8_t _CMD_CharIsNumeric(uint8_t c);
void _CMD_IncCursor();
void _CMD_DecCursor();
void _CMD_ClearCursor();
void _CMD_SetCursor(uint8_t x, uint8_t y);
void _CMD_ProcessControlChar(uint8_t c);
void _CMD_SwapCursorVals(uint8_t m);
void _CMD_LoadVersion();


// API Routines
void CMD_Init()
{
    // Figure out our position
    DISP_CFG_TRIS = 0x00;           // Default bits to output to keep them driven
    DISP_CFG_TRIS = DISP_CFG_MASK;  // Config bits are input
    DISP_CFG_ANSEL = 0x00;          // Digital
    _CMD_InitPosition();

    // State
    cmdState = CMD_STATE_IDLE;
    cmdDisplayMask = DISP_MEM_MASK_IMM;
    
    // Setup default cursor information
    cmdCursorEnable = 1;
    cmdCursorMode = CMD_CURSOR_MODE_BLINK;
    cmdCursorSaveCol[0] = 0;
    cmdCursorSaveCol[1] = 0;
    cmdCursorSaveRow[0] = 0;
    cmdCursorSaveRow[1] = 0;
    _CMD_SetCursor(0, 0);
}


void CMD_Tasks(uint8_t timerTick)
{
    uint8_t curCol;

    // Get the correct column for any displayed cursor
    if (cmdDisplayMask == DISP_MEM_MASK_IMM) {
        // Immediate Mode: Current operating cursor value
        curCol = cmdCursorCol;
    } else {
        // Shadow mode: Saved immediate cursor value
        curCol = cmdCursorSaveCol[0];
    }

    // Evaluate the cursor blink if necessary
    if ((timerTick == 1) &&
        (cmdCursorEnable == 1) &&
        (cmdCursorMode == CMD_CURSOR_MODE_BLINK) &&
        (cmdCursorState != CMD_CURSOR_STATE_NOTVISIBLE)) {

        if (cmdCursorState == CMD_CURSOR_STATE_VISIBLE_ON) {
            cmdCursorState = CMD_CURSOR_STATE_VISIBLE_OFF;
            DISP_SetCursor(_CMD_MyOffset(curCol), 0);
        } else {
            cmdCursorState = CMD_CURSOR_STATE_VISIBLE_ON;
            DISP_SetCursor(_CMD_MyOffset(curCol), 1);
        }
    }
}


void CMD_ProcessChar(uint8_t c)
{
    uint8_t ourPos = _CMD_PosIsOurs(cmdCursorCol, cmdCursorRow);

    switch (cmdState) {
        case CMD_STATE_IDLE:
            if ((c < 0x20) || (c == CMD_DEL)) {
                _CMD_ProcessControlChar(c);
            } else if (_CMD_CharIsDisplayable(&c)) {
                if (ourPos == 1) {
                    // Update our display
                    DISP_SetChar(_CMD_MyOffset(cmdCursorCol), c, cmdDisplayMask);
                }
                _CMD_IncCursor();
            }
            break;
        case CMD_STATE_CUR_COL:
            if (_CMD_CharIsNumeric(c)) {
                cmdCursorNewCol = (cmdCursorNewCol * 10) + (c - 0x30);
            } else if (c == ';') {
                cmdState = CMD_STATE_CUR_ROW;
            } else if (c == 'H') {
                _CMD_SetCursor(cmdCursorNewCol, cmdCursorNewRow);
                cmdState = CMD_STATE_IDLE;
            }
            break;
        case CMD_STATE_CUR_ROW:
            if (_CMD_CharIsNumeric(c)) {
                cmdCursorNewRow = (cmdCursorNewRow * 10) + (c - 0x30);
            } else if (c == 'H') { 
                _CMD_SetCursor(cmdCursorNewCol, cmdCursorNewRow);
                cmdState = CMD_STATE_IDLE;
            }
            break;
        default:
            cmdState = CMD_STATE_IDLE;
    }
}



// Internal Routines

void _CMD_InitPosition()
{
    uint8_t raw = ~DISP_CFG_PORT;  // Invert bit meanings (empty jumper = 0)

    cmdDisplayCol = (raw & DISP_CFG_COL_MASK) >> DISP_CFG_COL_SHIFT;
    cmdDisplayRow = (raw & DISP_CFG_ROW_MASK) >> DISP_CFG_ROW_SHIFT;
    cmdDisplayX = cmdDisplayCol * DISP_NUM_CHARS;
}


uint8_t _CMD_PosIsOurs(uint8_t x, uint8_t y)
{
    if (y != cmdDisplayRow) {
        return 0;
    } else if ((x >= cmdDisplayX) && (x < (cmdDisplayX + DISP_NUM_CHARS))) {
        return 1;
    } else {
        return 0;
    }
}


uint8_t _CMD_MyOffset(uint8_t x)
{
    return(x - cmdDisplayX);
}


uint8_t _CMD_CharIsDisplayable(uint8_t* c)
{
    if ((*c >= 0x20) && (*c <= 0x5F)) {
        // Directly displayable
        return 1;
    } else if (*c == 0x60) {
        // Convert Grave Accent to single quote
        *c = 0x27;
        return 1;
    } else if ((*c >= 0x61) && (*c <= 0x7F)) {
        // Convert lower case to upper case and {|}~ to []\_
        *c -= 0x20;
        return 1;
    } else {
        return 0;
    }
}


uint8_t _CMD_CharIsNumeric(uint8_t c)
{
    if ((c >= 0x30) && (c <= 0x39)) {
        return 1;
    } else {
        return 0;
    }
}


void _CMD_IncCursor()
{
    uint8_t newCol, newRow;
    
    // Increment the cursor position by 1
    if (cmdCursorCol < (MAX_COLS-1)) {
        // Bump to next location on this line
        newCol = cmdCursorCol + 1;
        newRow = cmdCursorRow;
    } else {
        if (cmdCursorRow < (MAX_ROWS-1)) {
            // Bump to start of next line
            newCol = 0;
            newRow = cmdCursorRow + 1;
        } else {
            // End of display: Bump one past end (and stay there)
            newCol = MAX_COLS;
            newRow = MAX_ROWS - 1;
        }
    }

    // Redisplay cursor in next location if possible
    _CMD_SetCursor(newCol, newRow);
}


void _CMD_DecCursor()
{
    uint8_t newCol, newRow;

    // Decrement the cursor position by 1
    if (cmdCursorCol == 0) {
        if (cmdCursorRow == 0) {
            // no change: Can't backspace past the start of the display
            newCol = 0;
            newRow = 0;
        } else {
            // End of previous row
            newCol = MAX_COLS - 1;
            newRow = cmdCursorRow - 1;
        }
    } else {
        // Previous location on this line
        newCol = cmdCursorCol - 1;
        newRow = cmdCursorRow;
    }

    // Redisplay cursor in next location if possible
    _CMD_SetCursor(newCol, newRow);
}


void _CMD_ClearCursor()
{
    uint8_t cursorVisible = (cmdCursorEnable == 1) &&
                            (cmdCursorState == CMD_CURSOR_STATE_VISIBLE_ON);

    // Remove the cursor from the current position if displayed
    if (cursorVisible) {
        DISP_SetCursor(_CMD_MyOffset(cmdCursorCol), 0);
        cmdCursorState = CMD_CURSOR_STATE_NOTVISIBLE;
    }
}


void _CMD_SetCursor(uint8_t x, uint8_t y)
{
    // Only change cursor state if we are in immediate mode
    if (cmdDisplayMask == DISP_MEM_MASK_IMM) {
        // Clear any previous cursor
        _CMD_ClearCursor();

        // Display cursor
        if ((cmdCursorEnable == 1) && (_CMD_PosIsOurs(x, y) == 1)) {
            cmdCursorState = CMD_CURSOR_STATE_VISIBLE_ON;
            DISP_SetCursor(_CMD_MyOffset(x), 1);
        } else {
            cmdCursorState = CMD_CURSOR_STATE_NOTVISIBLE;
        }
    }
    
    cmdCursorCol = x;
    cmdCursorRow = y;
}


void _CMD_ProcessControlChar(uint8_t c)
{
    switch (c) {
        case CMD_BS:
            _CMD_DecCursor();
            break;
        case CMD_TAB:
            _CMD_IncCursor();
            break;
        case CMD_LF:
            if (cmdCursorRow < (MAX_ROWS-1)) {
                // Move cursor to next line
                _CMD_SetCursor(cmdCursorCol, cmdCursorRow+1);
            }
            break;
        case CMD_FF:
            // Clear display
            DISP_Clear(cmdDisplayMask);
            _CMD_SetCursor(0, 0);
            break;
        case CMD_CR:
            // Move cursor to start of line
            _CMD_SetCursor(0, cmdCursorRow);
            break;
        case CMD_SO:
            // Setup Shadow
            if (cmdDisplayMask == DISP_MEM_MASK_IMM) {
                _CMD_SwapCursorVals(DISP_MEM_MASK_SHDW);
            }
            break;
        case CMD_SI:
            // Setup Immediate
            if (cmdDisplayMask == DISP_MEM_MASK_SHDW) {
                _CMD_SwapCursorVals(DISP_MEM_MASK_IMM);
            }
            break;
        case CMD_DC1:
            // Enable Cursor
            if (cmdDisplayMask == DISP_MEM_MASK_IMM) {
                if (cmdCursorEnable == 0) {
                    cmdCursorEnable = 1;
                    _CMD_SetCursor(cmdCursorCol, cmdCursorRow);
                }
            }
            break;
        case CMD_DC2:
            // Disable Cursor
            if (cmdDisplayMask == DISP_MEM_MASK_IMM) {
                _CMD_ClearCursor();
                cmdCursorEnable = 0;
            }
            break;
        case CMD_DC3:
            // Enable Cursor Blink
            if (cmdDisplayMask == DISP_MEM_MASK_IMM) {
                cmdCursorMode = CMD_CURSOR_MODE_BLINK;
            }
            break;
        case CMD_DC4:
            // Disable Cursor Blink
            if (cmdDisplayMask == DISP_MEM_MASK_IMM) {
                cmdCursorMode = CMD_CURSOR_MODE_SOLID;
                if ((cmdCursorState == CMD_CURSOR_STATE_VISIBLE_OFF) &&
                    (cmdCursorEnable == 1)) {

                    // Force cursor on if was in the "off" of a blink
                    cmdCursorState = CMD_CURSOR_STATE_VISIBLE_ON;
                    DISP_SetCursor(_CMD_MyOffset(cmdCursorCol), 1);
                }
            }
            break;
        case CMD_V:
            // Replace any current display with our version
            DISP_Clear(DISP_MEM_MASK_IMM);
            _CMD_LoadVersion();
            break;
        case CMD_SUB:
            // Load shadow memory to display
            DISP_LoadShadowMemory();

            // Setup immediate if necessary
            if (cmdDisplayMask == DISP_MEM_MASK_SHDW) {
                _CMD_SwapCursorVals(DISP_MEM_MASK_IMM);
            }
            break;
        case CMD_ESC:
            // Setup for a cursor update
            cmdCursorNewCol = 0;
            cmdCursorNewRow = 0;
            cmdState = CMD_STATE_CUR_COL;
            break;
        case CMD_DEL:
            // Backspace
            _CMD_DecCursor();

            // Delete character at new position
            if (_CMD_PosIsOurs(cmdCursorCol, cmdCursorRow) == 1) {
                DISP_SetChar(_CMD_MyOffset(cmdCursorCol), ' ', cmdDisplayMask);
            }
            break;
    }
}


void _CMD_SwapCursorVals(uint8_t m)
{
    if (m == DISP_MEM_MASK_IMM) {
        cmdCursorSaveCol[1] = cmdCursorCol;
        cmdCursorSaveRow[1] = cmdCursorRow;
        cmdCursorCol = cmdCursorSaveCol[0];
        cmdCursorRow = cmdCursorSaveRow[0];
    } else {
        cmdCursorSaveCol[0] = cmdCursorCol;
        cmdCursorSaveRow[0] = cmdCursorRow;
        cmdCursorCol = cmdCursorSaveCol[1];
        cmdCursorRow = cmdCursorSaveRow[1];
    }
    cmdDisplayMask = m;
}


void _CMD_LoadVersion()
{
    DISP_SetChar(0, 'V', DISP_MEM_MASK_IMM);

    DISP_SetChar(1, ((FW_MAJOR >> 4) & 0x0F) + '0', DISP_MEM_MASK_IMM);
    DISP_SetChar(2, (FW_MAJOR & 0x0F) + '0', DISP_MEM_MASK_IMM);

    DISP_SetChar(3, '.', DISP_MEM_MASK_IMM);

    DISP_SetChar(4, ((FW_MINOR >> 4) & 0x0F) + '0', DISP_MEM_MASK_IMM);
    DISP_SetChar(5, (FW_MINOR & 0x0F) + '0', DISP_MEM_MASK_IMM);
}