/********************************************************************
 * DL1416 Display control module
 *
 *
 *******************************************************************/
#include "display.h"
#include "system.h"
#include "usb_if.h"


// Internal variables
uint8_t disp_shadowMem[DISP_NUM_CHARS];
uint8_t disp_ShiftData;  // {unused, D6, D5, D4, D3, D2, D1, D0}
uint8_t disp_ShiftAddr;  // {unused, CU_N, CE3_N, CE2_N, CE1_N, CE0_N, A1, A0}


// Forward declarations
void _DISP_writeData();
uint8_t _DISP_posToCeMask(uint8_t pos);


// API Routines

void DISP_Init()
{
    uint8_t i;

    // Initialize the IO ports
    DISP_SCLK_TRIS = 0;
    DISP_SCLK_ANSEL = 0;
    DISP_SD0_TRIS = 0;
    DISP_SD0_ANSEL = 0;
    DISP_SD1_TRIS = 0;
    DISP_WRN_TRIS = 0;

    DISP_SCLK = 0;
    DISP_WRN = 1;

    // Clear the display
    DISP_Clear(DISP_MEM_MASK_BOTH);

    // Make sure there is no displayed cursor (which may occur at power-up)
    for (i=0; i<4; i++) {
        disp_ShiftData = 0;
        disp_ShiftAddr = i; // All four CEs enabled, cursor enabled
        _DISP_writeData();
    }
}


void DISP_Clear(uint8_t dispMemMask)
{
    uint8_t i;

    // Initialize the physical display if requested
    if (dispMemMask & DISP_MEM_MASK_IMM) {
        for (i=0; i<4; i++) {
            // Load a blank into the digit position in all four displays
            disp_ShiftData = DISP_BLANK;
            disp_ShiftAddr = 0x40 | i; // All four CEs enabled, cursor disabled
            _DISP_writeData();
        }
    }

    // Initialize the shadow RAM if requested
    if (dispMemMask & DISP_MEM_MASK_SHDW) {
        for (i=0; i<DISP_NUM_CHARS; i++) {
            disp_shadowMem[i] = DISP_BLANK;
        }
    }
}


void DISP_SetChar(uint8_t pos, uint8_t ch, uint8_t dispMemMask)
{
    // Only in legal positions
    if (pos >= DISP_NUM_CHARS)
        return;

    if (dispMemMask & DISP_MEM_MASK_IMM) {
        disp_ShiftAddr = 0x40 | _DISP_posToCeMask(pos) | (~pos & 0x03);
        disp_ShiftData = ch;
        _DISP_writeData();
    }

    if (dispMemMask & DISP_MEM_MASK_SHDW) {
        disp_shadowMem[pos] = ch;
    }
}


void DISP_SetCursor(uint8_t pos, uint8_t en)
{
    // Only in legal positions
    if (pos >= DISP_NUM_CHARS)
        return;

    disp_ShiftAddr = _DISP_posToCeMask(pos);  // CU_N = 0
    disp_ShiftData = (en) ? 0x01 << (~pos & 0x03) : 0x00;
    _DISP_writeData();
}


void DISP_LoadShadowMemory()
{
    uint8_t i;

    for (i=0; i<DISP_NUM_CHARS; i++) {
        disp_ShiftAddr = 0x40 | _DISP_posToCeMask(i) | (~i & 0x03);
        disp_ShiftData = disp_shadowMem[i];
        _DISP_writeData();
    }
}


// Internal Routines
void _DISP_writeData()
{
    uint8_t i;

    // Shift out the two control words MSbit first (they are destroyed)
    i = 8;
    while (i-- > 0) {
        DISP_SD0 = (disp_ShiftAddr & 0x80) ? 1 : 0;
        DISP_SD1 = (disp_ShiftData & 0x80) ? 1 : 0;
        disp_ShiftAddr <<= 1;
        disp_ShiftData <<= 1;
        DISP_SCLK = 1; // Minimum clock width is less than our instruction cycle
        //asm("NOP");
        DISP_SCLK = 0;
    }

    // Drive the write signal for at least 350 nSec (5 48 MHz cycles)
    asm("NOP");
    asm("NOP");
    DISP_WRN = 0;
    asm("NOP");
    asm("NOP");
    asm("NOP");
    asm("NOP");
    asm("NOP");
    DISP_WRN = 1;
}

uint8_t _DISP_posToCeMask(uint8_t pos)
{
    if (pos < 4)
        return 0x38;
    else if (pos < 8)
        return 0x34;
    else if (pos < 12)
        return 0x2C;
    else
        return 0x1C;
}
