/********************************************************************
 * 16-character DL1416 LED display program
 *
 * Features
 *   1. Supports 4 DL1416 displays using a 2-bit simplified SPI interface
 *      (designed to drive two 74HCT164 shift registers).
 *   2. Simple serial based command set.
 *   3. Dual communication interfaces: CDC-class USB or Serial
 *      Serial interface repeats incoming data stream to allow multiple
 *      displays.
 *   4. Supports up to 16x64 character display (4 column, 16 row addresses).
 *
 * Copyright (c) 2016 Dan Julio
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * So...if you come into a huge stash of DL1416 displays and decide to make
 * a bunch of these for a little pocket change, send me a few, or at least
 * credit me for the design.  Oh yeah, and set your own VID/PID in
 * usb_descriptors.c.  Thanks.  :-)
 * 
 *******************************************************************/

/** INCLUDES *******************************************************/
#include <xc.h>
#include <stdbool.h>
#include <stdint.h>
#include "cmd.h"
#include "display.h"
#include "sif.h"
#include "system.h"
#include "usb_if.h"
#include "usb.h"
#include "usb_device.h"
#include "usb_device_cdc.h"


/** CONFIGURATION Bits **********************************************/
// PIC16F1459 configuration bit settings:
    // CONFIG1
    #pragma config FOSC = INTOSC    // Oscillator Selection Bits (INTOSC oscillator: I/O function on CLKIN pin)
    #pragma config WDTE = OFF       // Watchdog Timer Enable (WDT disabled)
    #pragma config PWRTE = OFF      // Power-up Timer Enable (PWRT disabled)
    #pragma config MCLRE = OFF      // MCLR Pin Function Select (MCLR/VPP pin function is digital input)
    #pragma config CP = ON          // Flash Program Memory Code Protection (Program memory code protection is disabled)
    #pragma config BOREN = ON       // Brown-out Reset Enable (Brown-out Reset enabled)
    #pragma config CLKOUTEN = OFF   // Clock Out Enable (CLKOUT function is disabled. I/O or oscillator function on the CLKOUT pin)
    #pragma config IESO = OFF       // Internal/External Switchover Mode (Internal/External Switchover Mode is disabled)
    #pragma config FCMEN = OFF      // Fail-Safe Clock Monitor Enable (Fail-Safe Clock Monitor is disabled)

    // CONFIG2
    #pragma config WRT = OFF        // Flash Memory Self-Write Protection (Write protection off)
    #pragma config CPUDIV = NOCLKDIV// CPU System Clock Selection Bit (NO CPU system divide)
    #pragma config USBLSCLK = 48MHz // USB Low SPeed Clock Selection bit (System clock expects 48 MHz, FS/LS USB CLKENs divide-by is set to 8.)
    #pragma config PLLMULT = 3x     // PLL Multipler Selection Bit (3x Output Frequency Selected)
    #pragma config PLLEN = ENABLED  // PLL Enable Bit (3x or 4x PLL Enabled)
    #pragma config STVREN = ON      // Stack Overflow/Underflow Reset Enable (Stack Overflow or Underflow will cause a Reset)
    #pragma config BORV = LO        // Brown-out Reset Voltage Selection (Brown-out Reset Voltage (Vbor), low trip point selected.)
    #pragma config LPBOR = OFF      // Low-Power Brown Out Reset (Low-Power BOR is disabled)
    #pragma config LVP = OFF        // Low-Voltage Programming Enable (High-voltage on MCLR/VPP must be used for programming)



/*** TIMER0 Related *****************************************/
// FOSC/4 = 12 MHz, Prescale 256:1 ==> 21.333 uS/count  ==> 234 counts for 4.992 mSec
#define TMR0_TO         (256-234)
// Fast tick is approx 20 mSec
#define TMR0_MSEC       5
#define FAST_TICK_MSEC  20
#define TMR0_HALFSEC_COUNT  (500 / TMR0_MSEC)

volatile bit TMR0tick;   // Set in the ISR, consumed by SYSTEM_Tasks
bit FastTick;   // Set in SYSTEM_Tasks for consumption in the main loop
bit HalfSecTick;  // Set in SYSTEM_Tasks for consumption in main loop
volatile uint8_t TMR0div4tickCount;
uint8_t FastTickCount;
uint8_t HalfSecTickCount;




// Forward declarations
void SYS_Init();
void SYS_EvalTimer0();



void main(void)
{
    // Initialize
    SYS_Init();
    CMD_Init();
    DISP_Init();
    SIF_Init();
    USBIF_Init();

    // Loop processing any incoming commands
    while(1) {
        // Evaluate our tick timer first to set timing flags to be used during
        // this evaluation cycle
        SYS_EvalTimer0();

        // Evaluate polled USB tasks
        USBDeviceTasks();

        // Evaluate our modules
        CMD_Tasks((HalfSecTick == 1) ? 1 : 0);
        SIF_Tasks();
        USBIF_Tasks();
    }
}


void interrupt SYS_InterruptHigh(void)
{
    // TIMER0
    if (INTCONbits.T0IF == 1) {
        // Reload the timer and note we saw a 5 mSec tick
        TMR0 = TMR0_TO;
        INTCONbits.T0IF = 0;   // Clear the interrupt
        TMR0tick = 1;
    }

    // SIF RX
    if ((PIE1bits.RCIE == 1) && (PIR1bits.RCIF == 1)) {
        SIF_RxData();
    }
}


bool USER_USB_CALLBACK_EVENT_HANDLER(USB_EVENT event, void *pdata, uint16_t size)
{
    switch( (int) event )
    {
        case EVENT_TRANSFER:
            break;

        case EVENT_SOF:
            break;

        case EVENT_SUSPEND:
            break;

        case EVENT_RESUME:
            break;

        case EVENT_CONFIGURED:
            CDCInitEP();
            break;

        case EVENT_SET_DESCRIPTOR:
            break;

        case EVENT_EP0_REQUEST:
            /* We have received a non-standard USB request.  The HID driver
             * needs to check to see if the request was for it. */
            USBCheckCDCRequest();
            break;

        case EVENT_BUS_ERROR:
            break;

        case EVENT_TRANSFER_TERMINATED:
            break;

        default:
            break;
    }
    return true;
}


void SYS_Init()
{
    // Configure the processor
    OPTION_REG = 0x07;  // Prescaler to TIMER0, Int Clock, 1:256

    // Turn on active clock tuning for USB full speed operation
    OSCCON = 0xFC;  // HFINTOSC @ 16MHz, 3X PLL, PLL enabled

    // Active clock tuning enabled for USB
    ACTCON = 0x90;

    // Enable TIMER0 interrupts
    INTCONbits.TMR0IE = 1;
    INTCONbits.GIE = 1;
}




void SYS_EvalTimer0()
{
    // Tick bits will be set if necessary
    FastTick = 0;
    HalfSecTick = 0;

    if (TMR0tick == 1) {
        TMR0tick = 0;
        if ((++FastTickCount & 0x03) == 0x00) {
            FastTick = 1;
        }
        if (--HalfSecTickCount == 0) {
            HalfSecTickCount = TMR0_HALFSEC_COUNT;
            HalfSecTick = 1;
        }
    }
}