/********************************************************************
 * Serial Interface Module - process incoming data from the EUSART
 * Interface and pass character data to the command processor.
 *
*******************************************************************/
#include "cmd.h"
#include "sif.h"
#include "system.h"
#include <xc.h>
#include <stdbool.h>
#include <stdint.h>


// SPBRG calculation for 16-bit async baudrate with BRGH=1
#define SPBRG_INIT (48000000/SIF_BAUDRATE - 4)/4


// Module variables
uint8_t cmdSIFrxBuffer[SIF_RX_BUF_SIZE];
uint8_t cmdSIFrxPushI;
uint8_t cmdSIFrxPopI;
uint8_t cmdSIFrxNum;
uint8_t cmdSIFtxBuffer[SIF_TX_BUF_SIZE];
uint8_t cmdSIFtxPushI;
uint8_t cmdSIFtxPopI;
uint8_t cmdSIFtxNum;


// Internal routine forward declarations
void _SIF_IncRxIndex(uint8_t* index);
void _SIF_IncTxIndex(uint8_t* index);


// API Routines
void SIF_Init()
{
    // Initialize the SIF IO Ports
    SIF_RX_TRIS = 1;
    SIF_RX_ANSEL = 0;
    SIF_TX_TRIS = 0;

    // Initialize the EUSART
    TXSTA = 0x24; // TXEN + BRGH
    RCSTA = 0x90; // SPEN + CREN
    BAUDCON = 0x08;
    SPBRGL = SPBRG_INIT & 0xFF;
    SPBRGH = (SPBRG_INIT >> 8) & 0xFF;

    // Initialize module variables
    cmdSIFrxPushI = 0;
    cmdSIFrxPopI = 0;
    cmdSIFrxNum = 0;
    cmdSIFtxPushI = 0;
    cmdSIFtxPopI = 0;
    cmdSIFtxNum = 0;

    // Enable RX interrupts
    PIE1bits.RCIE = 1;
    INTCONbits.PEIE = 1;
    INTCONbits.GIE = 1;
}


void SIF_Tasks()
{
    uint8_t c;

    // Look for rx data to process
    if (cmdSIFrxNum > 0) {
        c = cmdSIFrxBuffer[cmdSIFrxPopI];
        _SIF_IncRxIndex(&cmdSIFrxPopI);
        cmdSIFrxNum--;
        CMD_ProcessChar(c);

        // Pass it to the next board
        SIF_PushTxChar(c);
    }

    // Look for tx data to transmit
    if (cmdSIFtxNum > 0) {
        if (PIR1bits.TXIF == 1) {
            TXREG = cmdSIFtxBuffer[cmdSIFtxPopI];
            _SIF_IncTxIndex(&cmdSIFtxPopI);
            cmdSIFtxNum--;
        }
    }
}


// Called by ISR
void SIF_RxData()
{
    cmdSIFrxBuffer[cmdSIFrxPushI] = RCREG;

    if (++cmdSIFrxPushI == SIF_RX_BUF_SIZE) {
        cmdSIFrxPushI = 0;
    }
    cmdSIFrxNum++;
}


void SIF_PushTxChar(uint8_t c)
{
    cmdSIFtxBuffer[cmdSIFtxPushI] = c;
    _SIF_IncTxIndex(&cmdSIFtxPushI);
    cmdSIFtxNum++;
}



// Internal Routines
void _SIF_IncRxIndex(uint8_t* index)
{
    if (++*index == SIF_RX_BUF_SIZE) {
        *index = 0;
    }
}


void _SIF_IncTxIndex(uint8_t* index)
{
    if (++*index == SIF_TX_BUF_SIZE) {
        *index = 0;
    }
}
